#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cstring>
#include <iomanip>
#include <cmath>
#include <ctime>
#include <bitset>
#include <vector>
#include <LL_global.h>

void Calc1dMoments(const std::vector<double>& a, const std::vector<double>& v, const std::vector<double>& delta_v, const double& m, double& n, double& v0, double& t);
void ReadInstrParameters(const std::string& filename, std::vector<unsigned long long>& start_times, std::vector<unsigned long long>& stop_times, std::vector<std::string>& ratio1_str, std::vector<std::string>& ratio2_str, std::vector<std::string>& spectra1_str, std::vector<std::string>& spectra2_str, std::vector<double>& m1, std::vector<double>& q1, std::vector<bool>& process1, std::vector<std::vector<double> >& eff1, std::vector<double>& m2, std::vector<double>& q2, std::vector<bool>& process2, std::vector<std::vector<double> >& eff2, std::vector<std::vector<double> >& geofactor, std::vector<std::vector<double> >& epq, std::vector<unsigned>& offset_ping, std::vector<unsigned>& offset_pong);
void RunCalculation(double Spectra[64], unsigned length, const std::vector<double>& eqtab, const std::vector<double>& eff, const std::vector<double>& geofactor, double m, double q, std::vector<double>& df, std::vector<double>& df_err, double& n, double& n_err, double& v, double& t);
void decompress(struct HIS_data* ll_packet);

void processHisData(struct HIS_data* ll_packet) {
	//decompress packet first
	decompress(ll_packet);


	// -- variables for holding intermediate results --
	std::vector<std::vector<double> > eqtab_vec, eff1_vec, eff2_vec, geofactor_vec;
	std::vector<unsigned> offset_ping_vec, offset_pong_vec;
	std::string instrinfo_filename = "his_extra_information.txt";

	std::vector<double> m1_vec;
	std::vector<double> q1_vec;
	std::vector<double> m2_vec;
	std::vector<double> q2_vec;

	std::vector<bool> process1_vec, process2_vec;
	std::vector<unsigned long long> start_times, stop_times;


	std::vector<std::string> ratio1_str_vec, ratio2_str_vec, spectra1_str_vec, spectra2_str_vec;

	// read from data file, either all of them with timerange or revision control

	ReadInstrParameters(instrinfo_filename, start_times, stop_times, ratio1_str_vec, ratio2_str_vec, spectra1_str_vec, spectra2_str_vec, m1_vec, q1_vec, process1_vec, eff1_vec, m2_vec, q2_vec, process2_vec, eff2_vec, geofactor_vec, eqtab_vec, offset_ping_vec, offset_pong_vec);

	int loc = -1;

	for (unsigned i = 0; i < start_times.size(); i++) {
		if ((ll_packet->SCET_seconds >= start_times[i]) && (ll_packet->SCET_seconds < stop_times[i])) {
			loc = i;
			break;
		}
	}

	if (loc == -1) {
		loc = start_times.size() - 1;
	}
	std::vector<double> eqtab = eqtab_vec[loc];
	std::vector<double> eff1 = eff1_vec[loc];
	std::vector<double> eff2 = eff2_vec[loc];
	std::vector<double> geofactor = geofactor_vec[loc];
	unsigned offset_ping = offset_ping_vec[loc];
	unsigned offset_pong = offset_pong_vec[loc];
	double m1 = m1_vec[loc];
	double q1 = q1_vec[loc];
	double m2 = m2_vec[loc];
	double q2 = q2_vec[loc];
	bool process1 = process1_vec[loc];
	bool process2 = process2_vec[loc];
	std::string ratio1_str = ratio1_str_vec[loc];
	std::string ratio2_str = ratio2_str_vec[loc];
	std::string spectra1_str = spectra1_str_vec[loc];
	std::string spectra2_str = spectra2_str_vec[loc];
	strcpy(ll_packet->ratio1_str, ratio1_str.c_str());
	strcpy(ll_packet->ratio2_str, ratio2_str.c_str());
	strcpy(ll_packet->spectra1_str, spectra1_str.c_str());
	strcpy(ll_packet->spectra2_str, spectra2_str.c_str());

	// define some notational short cuts
	std::vector<double> df1, df2;
	std::vector<double> df1_error, df2_error;
	double v1, v2;
	double n1, n2;
	double n1_error, n2_error;	
	double t1, t2;

	//check which kind of low latency it is
	if (ll_packet->structure_id > 132) {
		if (ll_packet->structure_id == 133) {
			std::vector<double> adjusted_spectra1;
			std::vector<double> adjusted_spectra2;
			adjusted_spectra1.resize(64, 0);
			adjusted_spectra2.resize(64, 0);
			for (unsigned i = 0; i < 32; i++) {
				adjusted_spectra1[offset_ping + i] = ll_packet->Spectra1[i];
				adjusted_spectra2[offset_ping + i] = ll_packet->Spectra2[i];
			}

			for (unsigned i = 0; i < 64; i++) {
				ll_packet->Spectra1[i] = adjusted_spectra1[i];
				ll_packet->Spectra2[i] = adjusted_spectra2[i];
			}				
		}

		if (ll_packet->structure_id == 134) {
			std::vector<double> adjusted_spectra1;
			std::vector<double> adjusted_spectra2;
			adjusted_spectra1.resize(64, 0);
			adjusted_spectra2.resize(64, 0);
			for (unsigned i = 0; i < 32; i++) {
				adjusted_spectra1[offset_pong + i] = ll_packet->Spectra1[i];
				adjusted_spectra2[offset_pong + i] = ll_packet->Spectra2[i];
			}

			for (unsigned i = 0; i < 64; i++) {
				ll_packet->Spectra1[i] = adjusted_spectra1[i];
				ll_packet->Spectra2[i] = adjusted_spectra2[i];
			}				
		}

	}


	if (process1) {
		RunCalculation(ll_packet->Spectra1, ll_packet->length, eqtab, eff1, geofactor, m1, q1, df1, df1_error, n1, n1_error, v1, t1);
	} else {
		df1.resize(64, -1E+31);
		df1_error.resize(64, -1E+31);
		v1 = -1E+31;
		n1 = -1E+31;
		n1_error = -1E+31;
		t1 = -1E+31;
	}

	if (process2) {
		RunCalculation(ll_packet->Spectra2, ll_packet->length, eqtab, eff2, geofactor, m2, q2, df2, df2_error, n2, n2_error, v2, t2);
	} else {
		df2.resize(64, -1E+31);
		df2_error.resize(64, -1E+31);
		v2 = -1E+31;
		n2 = -1E+31;
		n2_error = -1E+31;
		t2 = -1E+31;
	}

	ll_packet->n1 = n1;
	ll_packet->n1_err = n1_error;
	ll_packet->v1 = v1;
	ll_packet->t1 = t1;
	for (unsigned i = 0; i < 64; i++) ll_packet->df1[i] = df1[i];
	for (unsigned i = 0; i < 64; i++) ll_packet->df1_err[i] = df1_error[i];

	ll_packet->n2 = n2;
	ll_packet->n2_err = n2_error;
	ll_packet->v2 = v2;
	ll_packet->t2 = t2;
	for (unsigned i = 0; i < 64; i++) ll_packet->df2[i] = df2[i];
	for (unsigned i = 0; i < 64; i++) ll_packet->df2_err[i] = df2_error[i];

	for (unsigned i = 0; i < 64; i++) ll_packet->eqtab[i] = eqtab[i];
}


void decompress(struct HIS_data* ll_packet) {
	//perform lossy c decompression on packet
	for (unsigned i = 0; i < 64; i++) {
		unsigned val1 = (unsigned)ll_packet->Spectra1[i];
		unsigned val2 = (unsigned)ll_packet->Spectra2[i];

		if ((val1 == 0) && (val2 == 0)) continue;

		unsigned e1 = 0, m1 = 0, result1 = 0;
		unsigned e2 = 0, m2 = 0, result2 = 0;

		if((val1 & 0xC0) == 0xC0) {
			m1 = val1 & 0x07;
			e1 = (val1 >> 3) & 0x1F;
			result1  = (8+m1)*(unsigned)std::pow(2., e1-12);
		} else {
			m1 = val1 & 0x0F;
			e1 = (val1 >> 4) & 0x0F;

			if (e1 == 0) {
				result1 = m1;
			} else {
				result1 = (16 + m1) * (unsigned)std::pow(2., e1-1); 
			}
		}

		if((val2 & 0xC0) == 0xC0) {
			m2 = val2 & 0x07;
			e2 = (val2 >> 3) & 0x1F;
			result2  = (8+m2)*(unsigned)std::pow(2., e2-12);
		} else {
			m2 = val2 & 0x0F;
			e2 = (val2 >> 4) & 0x0F;

			if (e2 == 0) {
				result2 = m2;
			} else {
				result2 = (16 + m2) * (unsigned)std::pow(2., e2-1); 
			}
		}

		ll_packet->Spectra1[i] = result1;
		ll_packet->Spectra2[i] = result2;
	}
}

void RunCalculation(double Spectra[64], unsigned length, const std::vector<double>& eqtab, const std::vector<double>& efftab, const std::vector<double>& geofactor, double m, double q, std::vector<double>& df, std::vector<double>& df_err, double& n, double& n_err, double& v0, double& t) {
	//setup vector df and df err
	df.resize(64, 0);
	df_err.resize(64, 0);

	// setup vector a and a err
	std::vector<double> a;
	a.resize(64, 0);
	std::vector<double> a_err;
	a_err.resize(64, 0);

	/* delta v at step i is an ave. velocity at that step */
	std::vector <double> v;
	v.resize(64, 0);

	std::vector <double> delta_v; 
	delta_v.resize(64, 0);

	// E/q acceptance ranges
	std::vector<double> delta_eq;
	delta_eq.resize(64, 0);

	// calc. v[i] and  for all i, km/s
	// note: constant is unit conversion, 9.648e4 e amu km^2 / keV s^2
	for (unsigned nepq = 0; nepq < 64; nepq++){
		v[nepq] = sqrt( 2.0 * eqtab[nepq]* q/m * 9.648e4 );
	}

	// calculate phase space density (A[i]) for all i
	// A[i] = Nsi / (sqrt(delta_eq) tau g eta vi^4)
	// units: s^3/(cm^2 km^4)


	double sum_A = 0.;  // sum of A over nedb for normalization
	double sigsq_f1_dc = 0.;
	double f1_dc = 0.;
	double sigsq_f0_dc = 0.;
	double f0_dc = 0.;
	for (unsigned nepq = 0; nepq < 64; nepq++){
		// calc. delta_v[i] between each v[i]
		if (nepq == 63) {
			/* This is somewhat arbitrary but few if any counts should land here so it should not matter. */
			delta_v[nepq] = delta_v[nepq-1];
		}
		else {
			delta_v[nepq] = v[nepq] - v[nepq+1];
		}
		
		//WHAT IN THE WORLD IS FABS????
		delta_v[nepq] = fabs(delta_v[nepq]); // absolute value

		// Efficiency
		// WHAT WOULD PAPSVOLT BE IN THIS CASE??? //post acceleration voltage, pull from data file
		double PapsVolt = 25.0;
		double etot = q*(eqtab[nepq] + PapsVolt);
		double eff = efftab[nepq];

		if (eff >= 0.) {
			// distribution function in s^3/km^6
			// normal mode average accum time in seconds
			double accum_time = 4.05125;

			if (length == 32) accum_time = 4.603125;
			//WHAT IS THE POINT OF MULTIPLYING BY 10^10????
			//((10^2 cm/ m)^2 * (10^3 m / km)^2) = 10^10 cm^2 / km^2  used for unit change geofactor is in cm^2 
			a[nepq] = 2 * Spectra[nepq] * 1.e10 / (accum_time * eff * geofactor[nepq] * std::pow(v[nepq],4));

			// Error
			if (Spectra[nepq] > 0.) {
				a_err[nepq] = sqrt(Spectra[nepq]) / (Spectra[nepq]);

				// NOTE:  These formulas assume 1D moments used below
				// square of error in F1 (divided by constants (dc))
				sigsq_f1_dc += Spectra[nepq] * std::pow( delta_v[nepq] / std::pow(v[nepq],3) / eff ,2);

				// F1 (divided by constants (dc))
				f1_dc += Spectra[nepq] * delta_v[nepq] / std::pow(v[nepq],3) / eff;

				// square of error in F0 (divided by constants (dc))
				sigsq_f0_dc += Spectra[nepq]*std::pow( delta_v[nepq] / std::pow(v[nepq],4) / eff ,2);

				// F0 (divided by constants (dc))
				f0_dc += Spectra[nepq] * delta_v[nepq] / std::pow(v[nepq],4) / eff;
			} else {
				a_err[nepq] = 0.0;
			}
		} else {
			a[nepq] = 0.0; //FIX: is this what we want?
			a_err[nepq] = -1.0;
		}

		sum_A += a[nepq];
	}

	// calculate error in density
	n_err = sqrt(4. * sigsq_f1_dc / std::pow(f1_dc,2) +sigsq_f0_dc / std::pow(f0_dc,2));


	Calc1dMoments(a, v, delta_v, m, n, v0, t);

	for (unsigned i = 0; i < 64; i++) {
		//convert psd to diff flux using dJ/dE = 1.6022 * 10 ^ -20 * f * v^2 / m
		//include conversion from s^3 / (cm^2 * km^4) -> s^3 / m^6 which equal 10 ^ -8
		//include conversion from (km/s)^2 -> (m/s) ^ 2 which equals 10 ^ 6
		df[i] = (((1.6022 * std::pow(10.0, -20)) * a[i] * v[i] * v[i]) / m) * std::pow(10.0, -8) * std::pow(10.0, 6);
		df_err[i] = (sqrt(Spectra[i]) / Spectra[i]) * df[i]; 
	}

}


void Calc1dMoments(const std::vector<double>& a, const std::vector<double>& v, const std::vector<double>& delta_v, const double& m, double& n, double& v0, double& t) {
  // calculate 0th moment (sums over i)
  // F0/D =  n / v0^2
  // units: s^2/km^5
  double F0 = 0.0;
  for (unsigned epq = 0; epq < 64; epq++){
    F0 += a[epq] * delta_v[epq];
  }
  
  // calculate 1st moment (sums over i)
  // F1/D = n / v0
  // units: s/km^4
  double F1 = 0.0;
  for (unsigned epq = 0; epq < 64; epq++){
    F1 += a[epq] * v[epq] * delta_v[epq]; 
  }
  //cout<<"does this"<<endl;
  v0 = F1/F0;

  // calculate 2nd moment (sums over i)
  // F2 / D = n vth^2 / v0^2
  // units: 1/(km^3)
  double F2 = 0.0;
  for (unsigned epq = 0; epq < 64; epq++){
    F2 += a[epq] * std::pow((v[epq] - v0),2) * delta_v[epq];
  }
  //cout<<"does this"<<endl;
  // calculate state quantity: thermal velocity (vth) in km/s
  double vth = sqrt( F2 / F0 );

  // temperature = mass * vth^2 / (degf * k)
  // degf = degrees of freedom, 2 for now
  // k = boltzmann constant
  double degf = 2;
  double k = 1.38064852 * std::pow(10.0, -23.0);
  t = m * vth * vth / (degf * k);

  double DutyCycle = 1.0; //filler for now
  
  // calculate state quantity: density (n) in cm^-3
  // n =  F1^2 / (D F0 factor)
  // factor (1 km / 1.0e5 cm)^3 converts from km^-3 to cm^-3
  if (DutyCycle > 0.) {
    n = std::pow(F1,2.) / (DutyCycle * F0 * 1.0e15);
  }
  else {
    n = -1.0;
  }
}

void ReadInstrParameters(const std::string& filename, std::vector<unsigned long long>& start_times, std::vector<unsigned long long>& stop_times, std::vector<std::string>& ratio1_str, std::vector<std::string>& ratio2_str, std::vector<std::string>& spectra1_str, std::vector<std::string>& spectra2_str, std::vector<double>& m1_vec, std::vector<double>& q1_vec, std::vector<bool>& process1_vec, std::vector<std::vector<double> >& eff1_vec, std::vector<double>& m2_vec, std::vector<double>& q2_vec, std::vector<bool>& process2_vec, std::vector<std::vector<double> >& eff2_vec, std::vector<std::vector<double> >& geofactor_vec, std::vector<std::vector<double> >& eqtab_vec, std::vector<unsigned>& offset_ping, std::vector<unsigned>& offset_pong) {
	std::ifstream infile;
	std::string fill, name;
	double value, m, q;
	unsigned long long start, stop;
	unsigned u_value;
	bool trueorfalse;
	unsigned num_entries;

	infile.open(filename.c_str());
	if (!infile.good()) exit(-1);

	infile >> num_entries;
	std::getline(infile, fill);

	for (unsigned i = 0; i < num_entries; i++) {
		// get first time range
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		infile >> start >> stop;
		std::getline(infile, fill);
		start_times.push_back(start);
		stop_times.push_back(stop);

		// get name of first ratio
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::getline(infile, name);
		ratio1_str.push_back(name);

		// get name of second ratio
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::getline(infile, name);
		ratio2_str.push_back(name);


		// get name of first spectra
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::getline(infile, name);
		spectra1_str.push_back(name);


		// get first mass/charge of rate
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		infile >> m >> q;
		std::getline(infile, fill);
		m1_vec.push_back(m);
		q1_vec.push_back(q);
		
		// check to see if need to run moments calculation on first rate
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		infile >> trueorfalse;
		std::getline(infile, fill);
		process1_vec.push_back(trueorfalse);

		// get first rate efficiencies
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::vector<double> eff1;
		eff1.resize(64, 0);
		for (unsigned i = 0; i < 64; i++) {
			infile >> value;
			eff1[i] = value;
		}
		std::getline(infile, fill);
		eff1_vec.push_back(eff1);

		// get second name of rate
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::getline(infile, name);
		spectra2_str.push_back(name);

		// get second mass/charge of rate
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		infile >> m >> q;
		std::getline(infile, fill);
		m2_vec.push_back(m);
		q2_vec.push_back(q);
		
		// check to see if need to run moments calculation on second rate
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		infile >> trueorfalse;
		std::getline(infile, fill);
		process2_vec.push_back(trueorfalse);
		
		// get second rate efficiencies
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::vector<double> eff2;
		eff2.resize(64, 0);
		for (unsigned i = 0; i < 64; i++) {
			infile >> value;
			eff2[i] = value;
		}
		std::getline(infile, fill);
		eff2_vec.push_back(eff2);

		// get geometric factor
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::vector<double> geofactor;
		geofactor.resize(64, 0);
		for (unsigned i = 0; i < 64; i++) {
			infile >> value;
			geofactor[i] = value;
		}
		std::getline(infile, fill);
		geofactor_vec.push_back(geofactor);

		// get eqtab
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		std::vector<double> eqtab;
		eqtab.resize(64, 0);
		for (unsigned i = 0; i < 64; i++) {
			infile >> value;
			eqtab[i] = value;
		}
		std::getline(infile, fill);
		eqtab_vec.push_back(eqtab);

		// get burst mode ping
		if (!infile.good()) exit(-1);		
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		infile >> u_value;
		std::getline(infile, fill);
		offset_ping.push_back(u_value);

		// get burst mode pong
		if (!infile.good()) exit(-1);
		while (infile.peek() == '/') {
			std::getline(infile, fill);
		}
		infile >> u_value;
		std::getline(infile, fill);
		offset_pong.push_back(u_value);
	}

	infile.close();
}
